import logging
import queue
from threading import Thread
from typing import Any, Callable, ContextManager, Mapping, Optional, Sequence

log = logging.getLogger(__name__)
STOP = -1
STOP_TASK = (
    STOP,
    STOP,
    STOP,
)


class Queue(queue.Queue):
    def clear(self):
        with self.mutex:
            unfinished_tasks = self.unfinished_tasks - len(self.queue)
            if unfinished_tasks <= 0:
                if unfinished_tasks < 0:
                    raise ValueError("task_done() called too many times")
                self.all_tasks_done.notify_all()
            self.unfinished_tasks = unfinished_tasks
            self.queue.clear()
            self.not_full.notify_all()


class AsyncTaskManager(ContextManager):
    """
    Multithreaded task manager.
    Note that the manager itself is not thread safe - meaning it can be used only in a single thread.
    """

    def __init__(
        self,
        workers_amount: int = 5,
        name: str = None,
        collect_results: bool = True,
        iter_timeout: Optional[float] = None,
    ):
        super().__init__()
        self.name = name or "Task manager"
        self.collect_results = collect_results
        self.iter_timeout = iter_timeout
        self._task_queue: Queue = Queue()
        self._result_queue: Queue = Queue()
        self._total_task_count = 0
        self._workers_count = 0
        self.start_workers(workers_amount)

    def __iter__(self):
        return self

    def __next__(self):
        if self.is_work_done and self.is_empty:
            raise StopIteration()
        return self.get_result(timeout=self.iter_timeout)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__del__()

    def __del__(self):
        self.kill()

    @property
    def is_work_done(self) -> bool:
        """
        :return: True iff all tasks processing is done.
        """
        return self._task_queue.unfinished_tasks == 0

    @property
    def is_empty(self) -> bool:
        """
        :return: True iff all results fetching is done.
        """
        return self._result_queue.unfinished_tasks == 0

    @property
    def is_closed(self) -> bool:
        """
        :return: True iff all worker threads are closed.
        """
        return self._workers_count == 0

    @property
    def total_task_count(self) -> int:
        return self._total_task_count

    def start_workers(self, workers_amount: int):
        log.debug(f"Starting {workers_amount} workers.")
        for i in range(workers_amount):
            thread_name = f"{self.name} - worker {i + 1}"
            thread = Thread(target=self._work, daemon=True, name=thread_name)
            thread.start()
            self._workers_count += 1

    def add_task(self, func: Callable, args: Sequence = None, kwargs: Mapping = None):
        if self.is_closed:
            return
        args = tuple(args) if args else ()
        kwargs = kwargs or {}
        self._total_task_count += 1
        self._task_queue.put((func, args, kwargs))

    def _work(self):
        log.debug("Worker starting.")
        while True:
            func, args, kwargs = self._task_queue.get()
            if func == STOP:
                self._task_queue.task_done()
                break
            result = func(*args, **kwargs)
            if self.collect_results:
                self._result_queue.put(result)
            self._task_queue.task_done()
        log.debug("Worker done.")

    def get_result(self, timeout: Optional[float] = 3) -> Any:
        result = self._result_queue.get(block=True, timeout=timeout)
        self._result_queue.task_done()
        return result

    def stop_workers(self, amount: int = None):
        if amount is None:
            amount = self._workers_count
        log.debug(f"Stopping {amount} workers.")
        for i in range(amount):
            self._task_queue.put(STOP_TASK)
            self._workers_count -= 1

    def kill(self):
        log.debug(f"Removing all items from task queue (current size is {self._task_queue.qsize()})...")
        self._task_queue.clear()
        self.join()

    def join(self):
        log.debug("Joining all workers...")
        self._task_queue.join()  # Wait until all work is done
        self.stop_workers()
        self._task_queue.join()  # Wait until all workers are dead



from time import sleep

from the_spymaster_util.measure_time import MeasureTime

SLEEP_TIME_SEC = 0.1


def test_measure_time_is_a_context_manager():
    with MeasureTime() as measure:
        sleep(SLEEP_TIME_SEC)

    assert abs(SLEEP_TIME_SEC - measure.duration.total_seconds()) < 0.01



from datetime import timedelta
from time import time


class MeasureTime:
    def __init__(self):
        self.start = self.end = 0

    def __enter__(self):
        self.start = time()
        return self

    def __exit__(self, *args, **kwargs):
        self.end = time()

    @property
    def delta(self) -> float:
        return round(self.duration.total_seconds(), 3)

    @property
    def duration(self) -> timedelta:
        return timedelta(seconds=self.end - self.start)



# ORDER MATTERS!
from .config import LazyConfig  # noqa: F401
from .logging import get_logger, get_dict_config, wrap  # noqa: F401
from .context_thread import context_aware_thread_init  # noqa: F401
from .async_task_manager import AsyncTaskManager  # noqa: F401
from .http_client import BaseHttpClient, JsonType  # noqa: F401



import threading
from typing import Any, Callable, Iterable, Mapping, Optional

from the_spymaster_util import get_logger

log = get_logger(__name__)


# Hacky way to make all threads inherit log context from theirs parent thread


def _make_target_context_aware(target: Optional[Callable], parent_context: dict) -> Optional[Callable]:
    if not target:
        return None

    def wrapper(*args, **kwargs):
        try:
            log.update_context(**parent_context)
        except:  # noqa
            pass
        return target(*args, **kwargs)

    return wrapper


original_thread_init = threading.Thread.__init__  # type: ignore


def context_aware_thread_init(  # noqa
    self,
    group=None,
    target: Optional[Callable] = None,
    name: Optional[str] = None,
    args: Iterable[Any] = None,
    daemon: bool = False,
    kwargs: Optional[Mapping[str, Any]] = None,
):
    context_aware_target = _make_target_context_aware(target=target, parent_context=log.context)
    original_thread_init(
        self, group=group, target=context_aware_target, name=name, args=args or [], kwargs=kwargs, daemon=daemon
    )


threading.Thread.__init__ = context_aware_thread_init  # type: ignore



import json
import logging
from typing import Any, Callable, Dict, Optional, Union

import requests
from requests import Response
from requests.adapters import HTTPAdapter
from urllib3 import Retry

log = logging.getLogger(__name__)

CONTEXT_HEADER_KEY = "x-spymaster-context"
CONTEXT_ID_HEADER_KEY = "x-spymaster-context-id"

JsonType = Union[str, int, float, bool, list, Dict[str, Any], None]

DEFAULT_RETRY_STRATEGY = Retry(
    raise_on_status=False,
    total=4,
    backoff_factor=0.3,
    status_forcelist=[429, 500, 502, 503, 504],
    allowed_methods=["HEAD", "OPTIONS", "GET", "POST", "PUT", "DELETE"],
)


class BaseHttpClient:
    def __init__(self, base_url: str, retry_strategy: Optional[Retry] = DEFAULT_RETRY_STRATEGY):
        self.base_url = base_url
        self.session = requests.Session()
        self.set_retry_strategy(retry_strategy)

    def set_retry_strategy(self, retry_strategy: Optional[Retry]):
        retry_adapter = HTTPAdapter(max_retries=retry_strategy)
        self.session.mount("http://", retry_adapter)
        self.session.mount("https://", retry_adapter)

    def _http_call(self, endpoint: str, method: Callable, parse: bool = True, **kwargs) -> Union[dict, Response]:
        url = f"{self.base_url}/{endpoint}"
        headers = kwargs.pop("headers", None) or {}
        log_context = getattr(log, "context", None)
        if log_context:
            headers[CONTEXT_HEADER_KEY] = json.dumps(log_context)
        response = method(url, headers=headers, **kwargs)
        _log_data(url=url, response=response)
        response.raise_for_status()
        if parse:
            return response.json()
        return response

    def _get(self, endpoint: str, data: dict, **kwargs) -> dict:
        return self._http_call(endpoint=endpoint, method=self.session.get, params=data, **kwargs)  # type: ignore

    def _post(self, endpoint: str, data: dict, **kwargs) -> dict:
        return self._http_call(endpoint=endpoint, method=self.session.post, json=data, **kwargs)  # type: ignore


def _log_data(url: str, response: Response):
    try:
        data = response.json()
    except Exception:  # noqa
        data = response.content
    log.debug(
        f"Got status code {response.status_code}.",
        extra={"status_code": response.status_code, "url": url, "data": data},
    )



from setuptools import setup

setup(
    name="the-spymaster-util",
    version="1.11.2",
    description="Common utilities and infra for The Spymaster game.",
    author="Asaf Kali",
    author_email="akali93@gmail.com",
    url="https://github.com/asaf-kali/the-spymaster-util",
    install_requires=[
        "pydantic~=1.9",
        "dynaconf~=3.1",
        "boto3~=1.24",
        "ulid-py~=1.1",
        "requests~=2.28",
    ],
    include_package_data=True,
)
